/*
 * $Id: device_file.c,v 1.9 2002/02/02 21:16:13 mulix Exp $
 */ 


/* This code was initially based upon code from the book "Linux Device */
/* Drivers" by Alessandro Rubini and Jonathan Corbet, published by */
/* O'Reilly & Associates.*/

#include <linux/config.h>
#define __NO_VERSION__
#include <linux/module.h>       /* module things */
#include <linux/kernel.h>       /* printk() */
#include <linux/slab.h>         /* kmalloc() */
#include <linux/fs.h>           /* everything... */
#include <linux/errno.h>        /* error codes */
#include <linux/types.h>        /* size_t */
#include <linux/proc_fs.h>
#include <linux/fcntl.h>        /* O_ACCMODE */
#include <linux/spinlock.h>     /* spinlock */
#include <asm/uaccess.h>        /* copy_to/from_user */

#include "update_rules.h"
#include "rule_engine_public.h"
#include "device_file.h"
#include "request_buffer.h"
#include "response_buffer.h"
#include "module_interface.h"

#define SCT_MAJOR 0   /* dynamic major by default */

/* the major number we use */
int sct_major = SCT_MAJOR;

/* the device struct */
sct_dev_t sct_dev;

/* number of current *writers*. only one allowed! */
unsigned int sct_s_count = 0;

/* the lock used to protect the count */
spinlock_t sct_s_lock = SPIN_LOCK_UNLOCKED;

/* forward declarations */

void handle_request_packet(sct_dev_t* dev);

void build_response_packet(resp_buf_t resp, unsigned int packet_id, 
			   unsigned int result_code);
/*
 * Open and close
 */

int sct_open (struct inode *inode, struct file *filp)
{
	unsigned int open_mode = (filp->f_flags & O_ACCMODE);

	if ((open_mode == O_WRONLY) || (open_mode == O_RDWR))
	{
		spin_lock(&sct_s_lock);
	
		if (sct_s_count) {
			spin_unlock(&sct_s_lock);
			return -EBUSY; /* already open */
		}
		sct_s_count++;
		spin_unlock(&sct_s_lock);
	}
    
	/* and use filp->private_data to point to the device data */
	filp->private_data = &sct_dev;

	/* initialize the request/response buffers */
	req_clear(&sct_dev.req_buf);
	resp_clear(&sct_dev.resp_buf);
    
	MOD_INC_USE_COUNT;
	return 0;          /* success */
}

int sct_release (struct inode *inode, struct file *filp)
{
	unsigned int open_mode = (filp->f_flags & O_ACCMODE);
	if ((open_mode == O_WRONLY) || (open_mode == O_RDWR))
	{
		/* no need for locking, there can only be one writer */
		sct_s_count--;
	}

	MOD_DEC_USE_COUNT;
	return 0;
}


/*
 * Data management: read and write
 */

ssize_t 
sct_write(struct file *filp, const char *buf, size_t count,
	  loff_t *f_pos)
{
	sct_dev_t *dev = filp->private_data;
	ssize_t retval = -ENOMEM; /* our most likely error */

	if (down_interruptible (&dev->sem))
		return -ERESTARTSYS;

	if ((retval = req_add_chunk(&dev->req_buf, buf, count))) 
		goto add_chunk_error;
	
	if (req_has_whole_packet(&dev->req_buf) == 1) {
		printk(KERN_DEBUG "sct_write: whoohoo! got a whole packet!\n"); 
		handle_request_packet(dev);
	}

	up (&dev->sem);    
	return count;
	
 add_chunk_error:
	printk("sct_write: req_add_chunk returned '%d'\n", retval); 
	up (&dev->sem);
	return retval;
}

ssize_t 
sct_read (struct file *filp, char *buf, size_t count,
	  loff_t *f_pos)
{
	sct_dev_t *dev = filp->private_data;
	ssize_t retval = 0;
	
	if (down_interruptible (&dev->sem))
		return -ERESTARTSYS;
	
	count = resp_pop_chunk(&dev->resp_buf, buf, count);
	if (count < 0) {
		retval = -EINVAL;
		goto nothing;
	}
    
	up (&dev->sem);
	return count;
	
 nothing:
	up (&dev->sem);
	return retval;
}

/*
 * The fops
 */

struct file_operations sct_fops = {
	read: sct_read,
	write: sct_write, 
	open: sct_open,
	release: sct_release
};

/* set an int from this buffer. returns sizeof(int) */
static inline size_t set_int(unsigned char* buf, int val)
{
	memcpy(buf, &val, sizeof(val));
	return (sizeof(val));
}

/* set an int from this buffer. returns sizeof(int) */
static inline size_t get_int(unsigned char* buf, int* val)
{
	memcpy(val, buf, sizeof(*val));
	return (sizeof(*val));
    
}

void build_response_packet(resp_buf_t resp, unsigned int packet_id, 
			   unsigned int result_code)
{
	unsigned char* data;

	resp_clear(resp);

	/* cannot return NULL */
	data = resp_get_data(resp);
    
	data += set_int(data, sizeof(int)*3);
	data += set_int(data, packet_id);
	data += set_int(data, result_code);

	/* ugly breakage of modularity. to be purged later */
	resp_set_data_left(resp, sizeof(int)*3);
}


void handle_request_packet(sct_dev_t* dev)
{
	int rule_id;
	sct_rule_when_t when;
	int syscall_id;
	sct_action_t* a;
	sct_filter_node_t* f;
	int retval, packet_rc = 0;
	int mux_func_id = 0;
	unsigned char* data, *rule_buf;
	int dummy, packet_id, type_of_request;
	int tracking_rule_len;

	req_buf_t req = &dev->req_buf;
	resp_buf_t resp = &dev->resp_buf;

	/* parse the user space supplied buffer */
	/* we are guaranteesd the packet is valid */
	/* when we get called, since req_has_whole_packet */
	/* has validated it */
	data = req_get_data(req);
	data += get_int(data, &dummy); /* skip the size field */
	data += get_int(data, &packet_id); /* the packet id field */
	data += get_int(data, &type_of_request); /* the type of request field */
	data += get_int(data, &tracking_rule_len);

	/* and now, get out the entire rule buffer. we */
	/* play a dirty trick here, and zero out the magic */ 
	/* cookie, in order to null terminate the buffer */
	/* while avoiding a memory allocation */
	rule_buf = data;
	data += tracking_rule_len;
	data += set_int(data, 0);

	/* are sanity checks necessary here? tracker_build_rule */
	/* should perform them for us */
	if ((retval = tracker_build_rule(rule_buf, tracking_rule_len, 
					 &rule_id, &when, 
					 &syscall_id, &mux_func_id, 
					 &a, &f)) < 0) {
		printk("handle_request_packet: tracker_build_rule returned %d\n", retval); 
		packet_rc = CTL_TRACKER_EBADTRACKINGRULE;
		goto err_build_rule; 
	}
    
	if ((retval = sanity_check_user_input(syscall_id, mux_func_id, 
					      when)) < 0) {
		/* not good enough, but all sanity_check_user_input will */
		/* return to us is -EINVAL on error. it needs serious */
		/* surgery */
		printk("handle_request_packet: sanity_check returned %d\n", retval); 
		packet_rc = CTL_TRACKER_EBADTRACKINGRULE;
		goto err_failed_sanity_check;
	}

	printk(KERN_NOTICE "handle_request_packet: about to modify rules!\n");

	/* modify rules */
	switch (type_of_request) {
	case(CTL_TRACKER_CMD_ADD_RULE):
		retval = tracker_add_rule(rule_id, when,
					  syscall_id, mux_func_id,
					  f, a);
		break;
	
	case(CTL_TRACKER_CMD_DEL_RULE):
		retval = tracker_del_rule(rule_id, when,
					  syscall_id, mux_func_id);
		break;
	
	case(CTL_TRACKER_CMD_DEL_ALL_RULES_FOR_SYSCALL):
		retval = tracker_del_all_rules_for_syscall(syscall_id,
							   mux_func_id);
		break;
	
	case(CTL_TRACKER_CMD_DEL_ALL_MODULE_RULES):
		retval = tracker_del_all_module_rules();
		break;
	
	case(CTL_TRACKER_CMD_PRINT_ALL_RULES):
		retval = tracker_print_all_rules();
		break;
	
	default:
		printk(KERN_NOTICE "device_file.c:handle_packet(): invalid or unsupported "
		       "type of request - %d\n", type_of_request);
		packet_rc = CTL_TRACKER_EINVAL;
		goto err_invalid_request;
	}

	switch (retval) {
	case(-EINTR):
		packet_rc = CTL_TRACKER_EINTR;
		break;
	case(-ENOMEM):
		packet_rc = CTL_TRACKER_ENOMEM;
		break;
	case (0):
		packet_rc = CTL_TRACKER_SUCCESS;
		break;
	default: /* catch all */
		packet_rc = CTL_TRACKER_EINVAL;
	}

	/* build response packet */
    
 err_invalid_request:
 err_build_rule:
 err_failed_sanity_check:
	build_response_packet(resp, packet_id, packet_rc);
	/* and we're done! */
}
