/*
 * $Id: request_buffer.c,v 1.5 2002/02/02 21:16:13 mulix Exp $
 */
#include <linux/errno.h>
#include <asm/uaccess.h>

#include "request_buffer.h"
#include "module_interface.h"

/* add to the end of the request buffer this user supplied chunk,  */
/* with this size.  */
/* returns 0 on succes, a negative error on failure.  */
/* NOTE: chunk should point into user space memory, add_chunk will  */
/* take care of the copying */
int req_add_chunk(req_buf_t buf, const unsigned char* chunk, size_t sz)
{
	if (!buf)
		return -EINVAL;

	/* check that we have enough free space */
	if (sz > buf->free_space)
		return -ENOSPC; /* TODO: in the future, enlarge the buffer */
    
	if (copy_from_user(buf->whead, chunk, sz)) {
		return -EFAULT;
	}

	buf->whead += sz;
	buf->data_size += sz;
	buf->free_space -= sz;
    
	return 0;
}

/* does this request buffer contain a whole request packet, or not yet? */
int req_has_whole_packet(req_buf_t buf)
{
	unsigned int req_sz;
	unsigned int cookie = DEV_FILE_MAGIC_COOKIE;
	unsigned int packet_cookie;
	unsigned char* cookie_loc; 
    
	if (!buf)
		return -EINVAL;

	/* get the size of this request packet, which is the first 'int' in */
	/* the data buffer */
	memcpy(&req_sz, buf->data, sizeof(req_sz));
	
	/* now, sanity checks on req_sz */
	/* if it's bigger than the entire space we have, this packet is invalid */
	if (req_sz > REQ_ORIG_BUF_SIZE) {
		return -EINVAL;
	}

	/* we dont have quite enough data yet */
	if (req_sz > buf->data_size) {
		return 0; /* not a full packet yet */
	}
    
	/* ok, so req_sz must be smaller or equal to the data_size. cool! */
    
	/* let's check the last int from the end, if it's the magic cookie */
	/* we expect */

#ifdef SCT_MODULE_DEBUG
	printk(KERN_DEBUG "request packet dump: "); 
	for (cookie_loc = buf->data; cookie_loc < buf->data + req_sz; cookie_loc += sizeof(int))
		printk("%0x ", *(unsigned int*)cookie_loc);
	printk("\n"); 
#endif
	
	cookie_loc = &buf->data[req_sz - sizeof(cookie)]; 

	memcpy(&packet_cookie, cookie_loc, sizeof(cookie));
	if (packet_cookie != cookie) {
		printk("req_has_whole_packet got invalid cookie - wanted %0x, got %0x\n", 
		       cookie, packet_cookie); 
		return -EINVAL; 
	}
	
	/* whee, success */
	return 1;
}

/* get the request buffer's raw data */
unsigned char* req_get_data(req_buf_t buf)
{
	return buf->data;
}

/*get the size of the data */
size_t req_get_data_size(req_buf_t buf)
{
	return buf->data_size;
}

/* clear this request buffer */
void req_clear(req_buf_t buf)
{
	if (!buf)
		return;

	/* set the write head to point to the beginning of the data buf */
	buf->whead = buf->data;
	/* we have this much free space. note: buf->data is a char array. if 
	   we change it to char ptr, need to keep its size explicitly */
	buf->free_space = REQ_ORIG_BUF_SIZE;
	buf->data_size = 0;
	/* empty the buffer, just in case */
	memset(buf->whead, 0, buf->free_space);
}
