/*
 * $Id: var_callbacks.c,v 1.5 2002/01/20 22:20:00 mulix Exp $
 */ 

#include <linux/kernel.h>
#include <linux/sched.h>
#include <asm/current.h> /* for current */

#include "rule_engine_public.h"
#include "filter_vars_public.h"
#include "struct_handling.h"
#include "string_util.h"

/*
 * returns number of ancestors.
 * if vec isn't null, it also sets the processes nodes
 * for ancestors var.
 */

static int read_ancestors_into_vector(sct_filter_node_t *vec)
{
	struct task_struct *t;
	sct_filter_node_t *f;
	int rc = 0;
  
	read_lock(&tasklist_lock);
	for (t = current->p_pptr; t; t = t->p_pptr, rc++)
	{
		if (vec) {
			f = sct_make_filter_vector(2);
			if (f) {
				/* insert may fail, if number of ancestors   */
				/* grew while two read_ancestors_into_vector */
				/* calls.                                    */
				if (sct_insert_filter_node(vec, rc, f) == 1) {
					sct_insert_filter_node(f, 0, sct_make_long_val((long)t->pid));
					sct_insert_filter_node(f, 1, sct_make_string_val(t->comm));
				} else {
					sct_free_filter_node(f);
					/* we dont fail the call, because the */
					/* vector is still filled with valid  */
					/* inforamtion.                       */
					break;
				}
			} else {
				rc = 0;
				break;
			}
		}
	}
	read_unlock(&tasklist_lock);
	return rc;
}
                                                                        
/* variable calculation callbacks. */
sct_filter_node_t* proc_current_callback(sct_var_type_t type, void *data) {
	sct_filter_node_t *f = NULL;
    
	switch (type) {
	case VT_PROC_UID:
		f = sct_make_long_val((long)current->uid);
		break;
	case VT_PROC_EUID:
		f = sct_make_long_val((long)current->euid);
		break;
	case VT_PROC_SUID:
		f = sct_make_long_val((long)current->suid);
		break;
	case VT_PROC_GID:
		f = sct_make_long_val((long)current->gid);
		break;
	case VT_PROC_EGID:
		f = sct_make_long_val((long)current->egid);
		break;
	case VT_PROC_SGID:
		f = sct_make_long_val((long)current->sgid);
		break;
	case VT_PROC_PID:
		f = sct_make_long_val((long)current->pid);
		break;
	case VT_PROC_COMM:
		f = sct_make_string_val(current->comm);
		break;
	case VT_PROC_PPID:
	case VT_PROC_PCOMM:
	{
		sct_filter_node_t *ppid = NULL, *pcomm = NULL;
                                                                        
		read_lock(&tasklist_lock);
		if (current->p_pptr) {
			ppid = sct_make_long_val((long)current->p_pptr->pid);
			pcomm = sct_make_string_val(current->p_pptr->comm);
		}
		read_unlock(&tasklist_lock);
		if (type == VT_PROC_PPID) {
			f = ppid;
			set_var(VT_PROC_PCOMM, pcomm);
		} else {
			f = pcomm;
			set_var(VT_PROC_PPID, ppid);
		}
		break;
	}
	case VT_PROC_ANCESTORS:
	{
		int p;
                                                                         
		p = read_ancestors_into_vector(NULL);
		if (p && (f = sct_make_filter_vector(p))) {
			p = read_ancestors_into_vector(f);
			if (!p) {
				sct_free_filter_node(f);
				f = NULL;
			}
		}
	}
	default:
		f = NULL;
		break;
	}
	return f;
}

sct_filter_node_t* global_retval_callback(sct_var_type_t type, void *data)
{
	sct_filter_node_t *f = NULL;
    
	switch (type) {
	case VT_RETVAL:
		f = sct_make_int_val((int)data);
		break;
	default:
		f = NULL;
		break;
	}

	return f;
}

sct_filter_node_t* global_param_callback(sct_var_type_t type, int param_index, sct_data_type_t dtype, void *data) 
{
	sct_filter_node_t *f = NULL;
    
	switch (dtype) {
	case DT_CHAR:
		f = sct_make_char_val(*((char*)data));
		break;
	case DT_BOOL:
	case DT_INT:
		f = sct_make_int_val(*((int*)data));
		break;
	case DT_UINT:
		f = sct_make_uint_val(*((unsigned int*)data));
		break;
	case DT_SHORT:
		f = sct_make_short_val(*((short*)data));
		break;
	case DT_USHORT:
		f = sct_make_ushort_val(*((unsigned short*)data));
		break;
	case DT_LONG:
		f = sct_make_long_val(*((long*)data));
		break;
	case DT_ULONG:
		f = sct_make_ulong_val(*((unsigned long*)data));
		break;
	case DT_FLOAT:
		f = sct_make_ulong_val(*((float*)data));
		break;
	case DT_STRING: 
	{
		int retval = 0;
		char *string = NULL;
		retval = getstring((char*)data, &string);
		if (retval == 0)
			f = sct_make_string_val_without_copy(string);
		break;
	}
	default:
		f = NULL;
		break;
	}
    
	return f;
}

/* get a filter node with a struct's field parameter value */
sct_filter_node_t* global_struct_field_param_callback(sct_var_type_t type,
					    	  int param_index,
					    	  int field_num,
					    	  sct_data_type_t dtype,
					    	  void *data,
					    	  int struct_type)
{
	sct_filter_node_t* field = NULL;
	struct pdata {
		char* data;
		int struct_type;
	} *p_data;
		/* the type according to which the struct should be accessed. */
	int real_struct_type = struct_type;
	struct struct_data* struct_data = NULL;
	struct struct_data* static_struct_data = NULL;
	int do_type_cast = 0;

    	p_data = (struct pdata*)data;
    	if (struct_type == -1) {
		real_struct_type = p_data->struct_type;
    	}
	else {
		do_type_cast = 1;
		static_struct_data = get_struct_data(p_data->struct_type);
		if (!static_struct_data) {
	    		printk(KERN_NOTICE "sct: global_struct_field_param_callback: internal error: struct type '%d' unknown.\n", p_data->struct_type);
			goto global_struct_field_param_callback_fail;
		}
	}

	struct_data = get_struct_data(real_struct_type);
	if (!struct_data) {
	    	printk(KERN_NOTICE "sct: global_struct_field_param_callback: error: struct type '%d' unknown.\n", real_struct_type);
		goto global_struct_field_param_callback_fail;
	}
	if (do_type_cast && struct_data->struct_size > static_struct_data->struct_size) {
	    	printk(KERN_NOTICE "sct: global_struct_field_param_callback: bad cast from struct type '%d' (size %d) to struct type '%d' (size %d).\n",
		       p_data->struct_type,
		       static_struct_data->struct_size,
		       real_struct_type,
		       struct_data->struct_size);
		goto global_struct_field_param_callback_fail;
	}
	if (struct_data->offsets_vec[field_num] >= 0) {
		void* loc = p_data->data + struct_data->offsets_vec[field_num];
		sct_data_type_t field_dtype = struct_data->types_vec[field_num];

		field = global_param_callback(type, param_index, field_dtype, loc);
	}
	else {
	    printk(KERN_NOTICE "sct: global_struct_field_param_callback: offset for field '%d' (%ld) not supported (unmatchable field???)\n",
		   field_num,
		   struct_data->offsets_vec[field_num]);
	}

  global_struct_field_param_callback_fail:
    return field;
}
