#!/usr/bin/perl -w

my $verbose = 0;

use strict;

# testing of correctness of the rule engine.
# runs the list of test programs found in '$tests_dir', provided their names
# appear in '$rules_file'.
#
# Each program is compiled, and then its source is scaned for test data. Test
# data contains a list of rules, which are used to generate a rules file.
# This file is then checked using 'sct_config', and uploaded into the kernel
# module. Then, the test program is ran, while this perl script scans the log
# file for logging messages (as specified by the test's data), or until
# a timeout occurs.
#
# note: sct's kernel modules must be loaded _before_ running this program.
# note2: this program must be ran as 'root', as it injects rules into the
#        kernel module.
# note3: for info about how to write more tests, see '$tests_dir/README.tests'.

# constants.
my $RULE_FILE = "/tmp/sct_test_rules_$$.conf";
my $COMPILED_PROG = "sct_test_$$";
my $COMPILED_PROG_PATH = "/tmp/$COMPILED_PROG";
my $COMPILE_COMMAND = "gcc -o $COMPILED_PROG_PATH";
my $SCT_CONFIG = "../sct_config/sct_config";
my $TAIL_CMD = "/usr/bin/tail -0f";
my $LOGFILE = "/var/log/messages";

my $tests_dir = "./correctness_tests";
my $rules_file = $tests_dir . "/rules.conf";

my %rules = ();


# make sure we run as 'root'.
my $whoami = `whoami`;
chomp $whoami;
($whoami eq "root") ||
    die "$0 - must be run as 'root'.\n";

open(RULES, "$rules_file") ||
    die "Cannot open rules file '$rules_file' - $!\n";


while (<RULES>) {
    chomp;
    my $test_file = $_;
    my $test_file_path = "$tests_dir/$test_file";
    my $test_title = "";
    my $test_syscall = "";
    my $test_when = "";
    my @test_rules = ();
    my $test_expect = -1;
    my $test_expect_params = "";
    my $test_timeout = -1;
    my $test_index = 1;

    print "Test file '$test_file_path:\n";

    open(TEST, "$test_file_path") ||
	die "Cannot open test file '$test_file_path' - $!\n";

    # compile the test program.
    my $rc = system("$COMPILE_COMMAND $test_file_path");
    ($rc == 0) ||
	die "Failed compiling test file '$test_file_path' - $rc\n";

    while (<TEST>) {
	/TEST_TITLE\s*:\s*(.*)$/ && ($test_title = $1);
	/TEST_SYSCALL\s*:\s*(\S+)\s*$/ && ($test_syscall = $1);
	/TEST_WHEN\s*:\s*(before|after)\s*$/ && ($test_when = $1);
	/TEST_EXPECT\s*:\s*(\d+)\s*$/ && ($test_expect = $1);
	/TEST_EXPECT_PARAMS\s*:\s*(.*)\s*$/ && ($test_expect_params = $1);
	/TEST_TIMEOUT\s*:\s*(\d+)\s*$/ && ($test_timeout = $1);
	/TEST_RULE_(\d+)\s*:\s*(.*)$/ && ($test_rules[$1] = "$2");
	/TEST_END/ && do {
	    &perform_test($test_file,
			 $test_file_path,
			 $test_index,
			 $test_title,
			 $test_syscall,
			 $test_when,
			 $test_expect,
			 $test_expect_params,
			 $test_timeout,
			 \@test_rules);
    	    $test_title = "";
    	    $test_syscall = "";
    	    $test_when = "";
    	    @test_rules = ();
    	    $test_expect = -1;
    	    $test_expect_params = "";
    	    $test_timeout = -1;
	    $test_index++;
	};
    }
    close(TEST);

    unlink($COMPILED_PROG_PATH) ||
	die "Failed erasing test program '$COMPILED_PROG_PATH' - $!\n";
} # end while (<RULES>)

close(RULES);

exit 0;

sub perform_test
{
    my ($test_file, $test_file_path, $test_index, $test_title,
	$test_syscall, $test_when, $test_expect, $test_expect_params,
	$test_timeout, $test_rules_ref) = @_;

    # check no test parameters are missing.
    ($test_title ne "") ||
	die "Test file '$test_file_path' missing 'TEST_TITLE' in text $test_index\n";
    ($test_syscall ne "") ||
	die "Test file '$test_file_path' missing 'TEST_SYSCALL' in test '$test_title' (test # ${test_index})\n";
    ($test_when ne "") ||
	die "Test file '$test_file_path' missing 'TEST_WHEN' in test '$test_title' (test # ${test_index})\n";
    ($test_expect ne "") ||
	die "Test file '$test_file_path' missing 'TEST_EXPECT' in test '$test_title' (test # ${test_index})\n";
    ($test_expect_params ne "") ||
	die "Test file '$test_file_path' missing 'TEST_EXPECT_PARAMS' in test '$test_title' (test # ${test_index})\n";
    ($test_timeout ne "") ||
	die "Test file '$test_file_path' missing 'TEST_TIMEOUT' in test '$test_title' (test # ${test_index})\n";
    (!defined(@$test_rules_ref->[1]) || @$test_rules_ref->[1] eq "") &&
	die "Test file '$test_file_path' has no 'TEST_RULE_*' in test '$test_title' (test # ${test_index})\n";

    $| = 1;
    print "Test '$test_title'...";

    if ($verbose) {
	print ("Test file '$test_file':\n");
	print ("  syscall: '$test_syscall'\n");
	print ("  when: '$test_when'\n");
	print ("  expect: '$test_expect'\n");
	my $i;
	for ($i = 1; $i <= $#{@$test_rules_ref}; $i++) {
	    print ("  rule $i: '@$test_rules_ref->[$i]'\n");
	}
    }

    # prepare a rules file.
    open(RULE_FILE, ">$RULE_FILE") ||
	die "Cannot open rule file '$RULE_FILE' for writing - $!\n";
    my $i;
    print RULE_FILE <<EOM
        log_format {
            before {sct_test_$$: [%comm]: %sname(%params) (rule %ruleid)}
            after {sct_test_$$: [%comm]: %sname(%params) = %retval (rule %ruleid)}
        }

EOM
    ;
    for ($i=1; $i <= $#{@$test_rules_ref}; $i++) {
        print RULE_FILE <<EOM
        rule {
	    syscall_name: $test_syscall
            rule_name: rule_${test_syscall}_${i}
            when: $test_when
	    filter_expression { (@$test_rules_ref->[$i]) && COMM ~= "$COMPILED_PROG"}
	    action { type=LOG }
        }

EOM
        ;
    }
    close(RULE_FILE) ||
	die "Error closing rule file '$RULE_FILE' - $!\n";
    # check the config file, using sct_config.
    my $rc = system("$SCT_CONFIG check $RULE_FILE >/dev/null 2>&1");
    ($rc == 0) ||
	die "Error parsing rule file '$RULE_FILE' - $rc\n";

    # upload the rules.
    $rc = system("$SCT_CONFIG upload $RULE_FILE >/dev/null 2>&1");
    ($rc == 0) ||
	die "Error uploading rule file '$RULE_FILE' - $rc\n";
    sleep(1);

    # run the program in a different process.
    my $pid = fork();
    ($pid == -1) &&
	die "Failed fork() - $!\n";
    if ($pid == 0) { # child process.
	sleep(2);  # allow parent process to open the log file.
	system("$COMPILED_PROG_PATH");
	exit(0);
    }
    else {
	my $test_cmd_pid = -1;
	($test_cmd_pid = open(LOG, "$TAIL_CMD $LOGFILE|")) ||
	   die "Failed running '$TAIL_CMD $LOGFILE' - $!\n";
	my $rule_id = -1;
	my $had_timeout = 0;
	my $logline = "";
      logloop:
	while (1) {
	    # chunk of code copied from 'man perlfunc', 'alarm' fucntion.
	    # it reads one line from the log file, with a timeout.
            eval {
                local $SIG{ALRM} = sub { die "alarm\n" };       # NB \n required
                alarm $test_timeout;
	        $logline = <LOG>;
                alarm 0;
            };
            die if $@ && $@ ne "alarm\n";       # propagate errors
            if ($@) {
                # timed out
		$had_timeout = 1;
		last logloop;
            }
	    # end of chunk copied from 'man perlfunc'.

	    (defined($logline)) ||
		last logloop;
	    chomp $logline;
	    (print "LOG:$logline\n") if $verbose;
	    ($logline =~ /sct_test_$$:\s*\[.*${COMPILED_PROG}.*\]:\s*${test_syscall}\(.*${test_expect_params}.*\).*\(rule\s*(\d+)\)\s*$/) && do {
		$rule_id = $1;
		(print "LOG: found rule '$rule_id'\n") if $verbose;
		last logloop;
	    };
	}
	kill 15, $test_cmd_pid;
	close(LOG);
	waitpid($pid, 0);
	if ($rule_id == $test_expect) {
	    print "Ok\n";
	}
	else {
	    if ($had_timeout) {
	        print "Failed (timeout)\n";
	    }
	    else {
	        print "Failed\n";
	    }
	}
    }
    unlink($RULE_FILE) ||
	die "Failed erasing rule file '$RULE_FILE' - $!\n";
}
